{$A+,B-,D+,E-,F-,G+,I+,L+,N+,O-,P-,Q-,R-,S+,T-,V+,X+,Y+}
{$M 16384,0,655360}

{***************************************************************************}
{*                                                                         *}
{*  FreeType Glyph Viewer.                                                 *}
{*                                                                         *}
{*                                                                         *}
{*    This small program will load a TrueType font file and allow          *}
{*    you to view/scale/rotate its glyphs. Glyphs are in the order         *}
{*    found within the 'glyf' table.                                       *}
{*                                                                         *}
{*  This source code has been compiled and run under both Virtual Pascal   *}
{*  on OS/2 and Borland's BP7.                                             *}
{*                                                                         *}
{***************************************************************************}

program Convert;

uses Crt, TTDisp, TTTypes, TTMemory, TTCalc, TTTables, Raster, TTFile;

{&PMTYPE NOVIO}

label skip1;

Type Qbyte = array[0..3] of byte;

const
  Precis  = 64;
  Precis2 = Precis div 2;

  PrecisAux = 1024;

  Centre_X : int = 320;
  Centre_Y : int = 225;

  Profile_Buff_Size = 64000;

  font:array[0..9] of array[0..7] of byte =
         (
           ($1C,$22,$22,$22,$22,$22,$22,$1C),
           ($08,$18,$28,$08,$08,$08,$08,$08),
           ($1C,$22,$02,$02,$04,$08,$10,$3E),
           ($1C,$22,$02,$04,$02,$02,$22,$1C),
           ($08,$10,$20,$48,$7C,$08,$08,$08),
           ($3E,$20,$20,$3C,$02,$02,$22,$1C),
           ($1C,$22,$20,$3C,$22,$22,$22,$1C),
           ($3E,$22,$02,$02,$04,$04,$08,$08),
           ($1C,$22,$22,$1C,$22,$22,$22,$1C),
           ($1C,$22,$22,$1E,$02,$02,$22,$1C)
         );




var
  data:array[1..60000] of byte;


  w:char;
  Font_Buffer : PStorage;

  curGlyphContours : PGlyphContours;

  num_pts : word;
  num_ctr : word;

  glyfArray : word;

  epts_ctr : PShortArray;

  xCoord : PStorage;
  yCoord : PStorage;
  Flag   : PByteArray;

  ymin, ymax, xmax, xmin, xsize : longint;
  res,  resB                    : int;

  resR : real;

  resX, resY : real;

  LastX, LastY : FixedPoint;

  numPoints, numContours : int;

  curGlyph        : ^TGlyph;
  curGlyphContour : PGlyphContour;

  Bit : TRasterBlock;

  yCur : integer;

  ScXMax, ScYMax,
  CntX, CntY : Integer;

  Rotation : int;  (* Angle modulo 1024 *)


Procedure InitRows;
var
  i: integer;
  P: Pointer;
begin

  Bit.rows  := 450;
  Bit.cols  := 80;
  Bit.width := 640;
  Bit.flow  := TTFlowDown;
  Bit.size  := 80*450;

  GetMem( Bit.buffer, Bit.size );
  if Bit.buffer = NIL then
   begin
    Writeln('ERREUR:InitRows:Not enough memory to allocate BitMap');
    halt(1);
   end;

  { Note that the render pool should be allocated from the font pool }
  { for vairous debugging reasons, and because we're still in alpha, }
  { we don't do it yet..                                             }

  GetMem( P, Profile_Buff_Size );
  if P=nil then
   begin
    writeln('ERREUR:InitRows:Not enough memory to allocate render pool');
    Halt(2);
   end;

  InitRasterizer( Bit, P, Profile_Buff_Size );

  FillChar( Bit.Buffer^, Bit.Size, 0 );
end;


Procedure ClearData;
var i: integer;
begin
  FillChar( Bit.Buffer^, Bit.Size, 0 );

  FreeMem( XCoord, SizeOf(FixedPoint)*numPoints );
  FreeMem( YCoord, SizeOf(FixedPoint)*numPoints );

  FreeMem( Flag, numPoints );
end;


Function LoadTrueTypeChar( idx : integer ) : boolean;
var
  off    : longint;
  x, y   : Real;
  i, szp : integer;
  j      : word;
  c, ct  : byte;
  Gl     : TGlyph;
  EM     : Word;
  CR, SR : Real;

begin
  LoadtrueTypeChar:=FALSE;
  if (idx<0) or (idx>=Num_Glyphs) then exit;

  CurGlyph := @Glyphs^[idx];
  Gl       := CurGlyph^;

  numPoints        := Gl.numberOfPoints;
  numContours      := Gl.numberOfContours;
  curGlyphContours := Gl.Contours;

  GetMem( XCoord, SizeOf(Fixed)*numPoints );
  GetMem( YCoord, SizeOf(Fixed)*numPoints );
  GetMem( Flag, numPoints );

  xMin := Gl.xMin;
  xMax := Gl.xMax;
  yMin := Gl.yMin;
  yMax := Gl.yMax;

  EM := Font_Header^.UnitsPerEM;

  dec( xMax, xMin );
  dec( yMax, yMin );

  dec ( res );
  resR := res/EM/2;

  xmax := trunc( xmax * resR + 0.5 );
  ymax := trunc( ymax * resR + 0.5 );

  CR := Cos( Rotation*Pi/512 );
  SR := Sin( Rotation*Pi/512 );

  for j:=0 to numPoints-1 do
   begin

    x := Gl.Points^[j].x * ( res / EM );
    y := Gl.Points^[j].y * ( res / EM );

    off := Trunc( Precis*( CR*(x-xmax) + SR*(y-ymax) ) );

    XCoord^[j] := Precis*Centre_X + off;
    (*
    XCoord^[j] := Precis*( Centre_X + off div Precis ) + Precis2;
    *)

    off := Trunc( Precis*( - SR*(x-xmax) + CR*(y-ymax) ) );

    YCoord^[j] := Precis*Centre_Y + off;
    (*
    YCoord^[j] := Precis*( Centre_Y + off div Precis ) + Precis2;
    *)

    Flag^[j] := Gl.Points^[j].flag and 1;
   end;

  inc ( res );
  resR := 1/res;

  xsize := ( xmax + 7 ) div 8;

  LoadTrueTypeChar:=TRUE;
end;


function ConvertRaster : boolean;
var
  B : Array[0..128] of Integer;
  i : integer;
  G : TGlyphRecord;
begin

  for i := 0 to numContours-1 do
    B[i] := CurGlyphContours^[i].Finish;

  G.Outlines  := numContours;
  G.OutStarts := @B;
  G.Points    := numPoints;
  G.XCoord    := XCoord;
  G.YCoord    := YCoord;
  G.Flag      := Flag;

  ConvertRaster := RenderGlyph ( G, res, res );
end;




var i: integer;
    heure,
    min1,
    min2,
    sec1,
    sec2,
    cent1,
    cent2  :
{$IFDEF OS2}
    longint;
{$ELSE}
    word;
{$ENDIF}

    C : Char;

    Filename : String;

label Fin;

var
  Fail : Int;


function i2s(c:integer):string;
Var s:string;
Begin
     Str(c,s);
     i2s:=s;
End;

procedure show(c:integer);
var p:integer;
    n:byte;
Begin
     p:=76;
     repeat
           n:=c mod 10;
           c:=c div 10;

           vio^[p+0*80]:=font[n][0];
           vio^[p+1*80]:=font[n][1];
           vio^[p+2*80]:=font[n][2];
           vio^[p+3*80]:=font[n][3];
           vio^[p+4*80]:=font[n][4];
           vio^[p+5*80]:=font[n][5];
           vio^[p+6*80]:=font[n][6];
           vio^[p+7*80]:=font[n][7];

           dec(p);
     until c=0;
End;


Function NoSpace(s:string):string;
Begin
     If Pos(' ',s)=0 then NoSpace:=s else NoSpace:=Copy(s,1,Pos(' ',s)-1);
End;

Function StrUpper(s:string):string;
Var i:byte;
    s1:string;
Begin
     s1:='';
     For i:=1 to length(s) do s1:=s1+upcase(s[i]);
     StrUpper:=s1;
End;

Function I2Slen(c:longint; l:byte):string;
Var i:byte;
    s1:string;
Begin
     Str(c:l,s1);
     For i:=1 to l do
      If s1[i]=' ' then s1[i]:='0';
     I2Slen:=s1;
End;

function Load_TrueType_Tables2 : Boolean;
var
  T : LongInt;
  L : LongInt;
  req,max_offset,max_len,sum_len,sum_len2:longint;
  i:integer;

  TableDir:TTableDir;
  TableDir_Entries:PTableDirEntries;

  name:string;
  f,fontfile:file;

begin
  Load_TrueType_Tables2 := False;

  if not Read_At_Font_File( 0, TableDir, sizeof(TableDir) ) then exit;

  Do32( Fixed( TableDir.Version ) );
  Do16( TableDir.NumTables );

{$IFDEF DEBUG}
{  Writeln('Version de rpertoire : ',TableDir.version/$10000);}
  Writeln('Nombre de Tables      : ',TableDir.numTables);

{$ENDIF}

  Num_TDE := TableDir.NumTables;

  L := sizeof(TTableDirEntry) * Num_TDE;

  if not Alloc( L, Pointer(TableDir_Entries) ) or
     not Read_Font_File( TableDir_Entries^, L ) then exit;

  max_offset:=0;
  max_len:=0;
  sum_len:=0;
  sum_len2:=0;

  writeln('Font tables:');
  writeln('name','offset':10,'length':10,'len-4b':10);
  writeln('------------------------------------------------------------------------------');

  for t:=0 to Num_TDE-1 do with TableDir_Entries^[t] do
    begin
{     CheckSum:=0;}
     Do32( Offset );
     Do32( Length );


     write(   TableDir_Entries^[t].Tag[0],
              TableDir_Entries^[t].Tag[1],
              TableDir_Entries^[t].Tag[2],
              TableDir_Entries^[t].Tag[3],
              TableDir_Entries^[t].Offset:10,
              TableDir_Entries^[t].Length:10);

     For i:=0 to 3 do
      If (TableDir_Entries^[t].length+i) mod 4 = 0 then
      Begin
           writeln(TableDir_Entries^[t].length+i:10);
           inc(sum_len,TableDir_Entries^[t].length+i);
      End;

     inc(sum_len2,TableDir_Entries^[t].length);

     If offset>max_offset then begin max_offset:=offset; max_len:=length; end;
   end;

  writeln('------------------------------------------------------------------------------');
  writeln(sum_len2:24,sum_len:10);
  writeln('+252':24,'+252':10);
  writeln('--------':24,'--------':10);
  writeln(sum_len2+252:24,sum_len+252:10);
  writeln;


  write('Writing "TABDIR.TTT" ...');
  assign(f,'tabdir.ttt');
  rewrite(f,1);
  blockwrite(f,TableDir,sizeof(TableDir));
  For t:=0 to Num_TDE-1 do BlockWrite(f,TableDir_Entries^[t],sizeof(TableDir_Entries^[t]));
  close(f);
  writeln(' done.');


  For t:=0 to Num_TDE-1 do   { nacitanie tabuliek }
  Begin
       name:= TableDir_Entries^[t].Tag[0]+
              TableDir_Entries^[t].Tag[1]+
              TableDir_Entries^[t].Tag[2]+
              TableDir_Entries^[t].Tag[3];

       write('Loading ',name,' table ...');

       If TableDir_Entries^[t].length>60000 then
       begin
            writeln('File too large !');
            halt;
       end;


       Read_At_Font_File( TableDir_Entries^[t].offset , data , TableDir_Entries^[t].length );

       assign(f,'table'+I2Slen(t,2)+'.ttt');
       rewrite(f,1);

       write(' done, writing to "TABLE',I2Slen(t,2),'.TTT" ...');
       BlockWrite(f,data,TableDir_Entries^[t].length);
       close(f);

       writeln(' done.');
  End;

  writeln;
  writeln('Now trying to reconstruct font to file "CONVERT.TTF" ...');

  assign(f,'tabdir.ttt');
  reset(f,1);

  assign(fontfile,'convert.ttf');
  rewrite(fontfile,1);

  blockread(f,TableDir,sizeof(TableDir));

  For t:=0 to TableDir.NumTables-1 do
   BlockRead(f,TableDir_Entries^[t],sizeof(TableDir_Entries^[t]));
  close(f);
  writeln('File "TABDIR.TTT" loaded.');

  seek(fontfile,12+16*TableDir.NumTables);

  For t:=0 to TableDir.NumTables-1 do
  Begin
       fillchar(data,sizeof(data),0);

       assign(f,'table'+I2Slen(t,2)+'.ttt');
       reset(f,1);
       blockread(f,data,TableDir_Entries^[t].length);
       close(f);

       for i:=0 to 3 do
        If (TableDir_Entries^[t].length+i) mod 4 = 0 then req:=i;

       seek(fontfile,TableDir_Entries^[t].offset);
       blockwrite(fontfile,data,req+TableDir_Entries^[t].length);

       name:= TableDir_Entries^[t].Tag[0]+
              TableDir_Entries^[t].Tag[1]+
              TableDir_Entries^[t].Tag[2]+
              TableDir_Entries^[t].Tag[3];

       writeln('Table ',name,' loaded.');
  End;

  seek(fontfile,12);

  For t:=0 to TableDir.NumTables-1 do
  Begin
       Do32(TableDir_Entries^[t].offset);
       Do32(TableDir_Entries^[t].length);

       BlockWrite(fontfile,TableDir_Entries^[t],sizeof(TableDir_Entries^[t]));
  End;
  writeln('Tables written.');

  Do32( Fixed( TableDir.Version ) );
  Do16( TableDir.NumTables );

  seek(fontfile,0);
  BlockWrite(fontfile,TableDir,sizeof(Tabledir));

  close(fontfile);
  writeln('Font succesfuly writen.');

  Load_TrueType_Tables2 := True;
end;


function TrueType_create_TTT : Boolean;
var
  T : LongInt;
  L : LongInt;
  req,max_offset,max_len,sum_len,sum_len2:longint;
  i:integer;

  TableDir:TTableDir;
  TableDir_Entries:PTableDirEntries;

  name:string;
  f,fontfile:file;

  file_text:text;

begin
  TrueType_create_TTT := False;

  if not Read_At_Font_File( 0, TableDir, sizeof(TableDir) ) then exit;

  Do32( Fixed( TableDir.Version ) );
  Do16( TableDir.NumTables );

{$IFDEF DEBUG}
{  Writeln('Version de rpertoire : ',TableDir.version/$10000);}
  Writeln('Nombre de Tables      : ',TableDir.numTables);

{$ENDIF}

  Num_TDE := TableDir.NumTables;

  L := sizeof(TTableDirEntry) * Num_TDE;

  if not Alloc( L, Pointer(TableDir_Entries) ) or
     not Read_Font_File( TableDir_Entries^, L ) then exit;

  max_offset:=0;
  max_len:=0;
  sum_len:=0;
  sum_len2:=0;

  writeln('Font tables:');
  writeln('name','offset':10,'length':10,'len-4b':10);
  writeln('------------------------------------------------------------------------------');

  for t:=0 to Num_TDE-1 do with TableDir_Entries^[t] do
    begin
     Do32( CheckSum );
     Do32( Offset );
     Do32( Length );


     write(   TableDir_Entries^[t].Tag[0],
              TableDir_Entries^[t].Tag[1],
              TableDir_Entries^[t].Tag[2],
              TableDir_Entries^[t].Tag[3],
              TableDir_Entries^[t].Offset:10,
              TableDir_Entries^[t].Length:10);

     For i:=0 to 3 do
      If (TableDir_Entries^[t].length+i) mod 4 = 0 then
      Begin
           writeln(TableDir_Entries^[t].length+i:10);
           inc(sum_len,TableDir_Entries^[t].length+i);
      End;

     inc(sum_len2,TableDir_Entries^[t].length);

     If offset>max_offset then begin max_offset:=offset; max_len:=length; end;
   end;

  writeln('------------------------------------------------------------------------------');
  writeln(sum_len2:24,sum_len:10);
  writeln('+252':24,'+252':10);
  writeln('--------':24,'--------':10);
  writeln(sum_len2+252:24,sum_len+252:10);
  writeln;


  write('Writing "TABDIR.TTT" ...');

  assign(file_text,'tabdir.ttt');
  rewrite(file_text);
  For t:=0 to Num_TDE-1 do
  Begin
       Writeln(file_text, TableDir_Entries^[t].Tag[0],
                          TableDir_Entries^[t].Tag[1],
                          TableDir_Entries^[t].Tag[2],
                          TableDir_Entries^[t].Tag[3] );
  End;
  close(file_text);


  assign(f,'header.ttt');
  rewrite(f,1);
  blockwrite(f,TableDir,sizeof(TableDir));
  {For t:=0 to Num_TDE-1 do BlockWrite(f,TableDir_Entries^[t],sizeof(TableDir_Entries^[t]));}
  close(f);
  writeln(' done.');


  For t:=0 to Num_TDE-1 do   { nacitanie tabuliek }
  Begin
       name:= TableDir_Entries^[t].Tag[0]+
              TableDir_Entries^[t].Tag[1]+
              TableDir_Entries^[t].Tag[2]+
              TableDir_Entries^[t].Tag[3];

       write('Loading ',name,' table ...');

       If TableDir_Entries^[t].length>60000 then
       begin
            writeln('File too large !');
            halt;
       end;


       Read_At_Font_File( TableDir_Entries^[t].offset , data , TableDir_Entries^[t].length );

       assign(f,'table'+I2Slen(t,2)+'.ttt');
       rewrite(f,1);

       write(' done, writing to "TABLE',I2Slen(t,2),'.TTT" ...');
       BlockWrite(f,data,TableDir_Entries^[t].length);
       close(f);

       writeln(' done.');
  End;

  TrueType_create_TTT:=true;

End;


Procedure Add32(var dest:qbyte; source:qbyte);
Var p:byte;
    i:byte;
    a:byte;
Begin
     p:=0;
     For i:=3 downto 0 do
     Begin
          a:=(word(dest[i])+word(source[i])+word(p)) and 255;
          p:=((word(dest[i])+word(source[i])+word(p)) shr 8) and 255;
          dest[i]:=a;
     End;
End;

function TrueType_create_TTF : Boolean;
var
  T : LongInt;
  L,free_ofs : LongInt;
  req,max_offset,max_len,sum_len,sum_len2:longint;

  i,ok_i:integer;

  TableDir:TTableDir;
  TableDir_Entries:PTableDirEntries;

  name:string;
  f,fontfile:file;

  file_text:text;

  sum,a:qbyte;

begin
  TrueType_create_TTF:=false;

  writeln;
  writeln('Now trying to reconstruct font to file "CONVERT.TTF" ...');

  assign(f,'header.ttt');
  reset(f,1);
  Blockread(f,TableDir,sizeof(TableDir));
  close(f);

  assign(file_text,'tabdir.ttt');
  reset(file_text);

  free_ofs:=12+16*TableDir.NumTables;

  For t:=0 to TableDir.NumTables-1 do
  Begin
       readln(file_text,name);

       TableDir_Entries^[t].Tag[0]:=name[1];
       TableDir_Entries^[t].Tag[1]:=name[2];
       TableDir_Entries^[t].Tag[2]:=name[3];
       TableDir_Entries^[t].Tag[3]:=name[4];

       assign(f,'table'+I2Slen(t,2)+'.ttt');
       reset(f,1);

       TableDir_Entries^[t].length:=filesize(f);
       TableDir_Entries^[t].offset:=free_ofs;

       inc(free_ofs,filesize(f));

       For i:=0 to 3 do
        If (free_ofs+i) mod 4 = 0 then ok_i:=i;

       inc(free_ofs,ok_i);

       close(f);
  End;

  close(file_text);

  writeln('File "TABDIR.TTT" loaded.');


{}

  assign(fontfile,paramstr(1));
  rewrite(fontfile,1);

  seek(fontfile,12+16*TableDir.NumTables);

  For t:=0 to TableDir.NumTables-1 do
  Begin
       fillchar(data,sizeof(data),0);

       assign(f,'table'+I2Slen(t,2)+'.ttt');
       reset(f,1);
       blockread(f,data,TableDir_Entries^[t].length);
       close(f);

       { checksum }
       fillchar(sum,4,0);
       For i:=1 to (TableDir_Entries^[t].length+3) div sizeof(ulong) do
       Begin
            Move(data[(i-1)*sizeof(ulong)+1],a,sizeof(ulong));
            add32(sum,a);
       End;
       { /checksum }
       Move(sum,TableDir_Entries^[t].checksum,4);


       for i:=0 to 3 do
        If (TableDir_Entries^[t].length+i) mod 4 = 0 then req:=i;

       seek(fontfile,TableDir_Entries^[t].offset);
       blockwrite(fontfile,data,req+TableDir_Entries^[t].length);

       name:= TableDir_Entries^[t].Tag[0]+
              TableDir_Entries^[t].Tag[1]+
              TableDir_Entries^[t].Tag[2]+
              TableDir_Entries^[t].Tag[3];

       writeln('Table ',name,' loaded, written at position ',TableDir_Entries^[t].offset,', size ',
                TableDir_Entries^[t].length);
  End;

  seek(fontfile,12);

  Do32(TableDir_Entries^[0].checksum);

  For t:=0 to TableDir.NumTables-1 do
  Begin
{       Do32(TableDir_Entries^[t].checksum);}
       Do32(TableDir_Entries^[t].offset);
       Do32(TableDir_Entries^[t].length);

       BlockWrite(fontfile,TableDir_Entries^[t],sizeof(TableDir_Entries^[t]));
  End;
  writeln('Tables written.');

  Do32( Fixed( TableDir.Version ) );
  Do16( TableDir.NumTables );

  seek(fontfile,0);
  BlockWrite(fontfile,TableDir,sizeof(Tabledir));

  close(fontfile);
  writeln('Font succesfuly writen.');

  TrueType_create_TTF := True;
end;



begin

  GetMem       ( Font_Buffer, 64000 );
  Init_FontPool( Font_Buffer^, 64000 );

  curGlyphContours:=NIL;

  num_pts   :=0;
  num_ctr   :=0;

  xCoord  :=NIL;
  yCoord  :=NIL;
  Flag    :=NIL;

  If paramCount<>1 then
   begin
    Writeln('Simple TrueType Glyphs viewer - part of the FreeType project' );
    Writeln;
    Writeln('Usage : ',paramStr(0),' FontName[.TTF]');
    Halt(1);
   end;

  write('[R]ead font to TABDIR.TTT / [W]rite font from TABDIR.TTT / ESC : ');
  Repeat
        w:=upcase(readkey);
  Until ((w='R') or (w='W') or (w=#27));

  If w='W' then goto skip1;

  Filename := ParamStr(1);
  if Pos('.',FileName)=0 then FileName:=FileName+'.TTF';
  if not Open_TrueType_File(Filename ) then
   begin
    Writeln('Erreur, le fichier ',ParamStr(1),' n''a pu tre ouvert');
    Halt(1);
   end;


skip1:


  Case w of
    'R' : TrueType_create_TTT;
    'W' : TrueType_create_TTF;
  End;

  If w='R' then Close_TrueType_File;
end.

