Unit TTTables;

interface

uses FreeType, TTTypes, TTCalc,crt;

{*****************************************************}
{*                                                   *}
{*              TrueType Tables Types                *}
{*                                                   *}
{*****************************************************}

{$I compdef.inc}

type
  {* Graphics State                            *}
  {*                                           *}
  {* The Graphics State (GS) is managed by the *}
  {* instruction field, but does not come from *}
  {* the font file. Thus, we can use 'int's    *}
  {* where needed.                             *}
  {*                                           *}

  PGraphicsState = ^TGraphicsState;
  TGraphicsState = record
                     autoFlip                : boolean;
                     controlValueCutIn       : TT_F26dot6;
                     deltaBase               : int;
                     deltaShift              : int;

                     dualVector,
                     projVector,
                     freeVector              : TT_UnitVector;

                     gep0,
                     gep1,
                     gep2                    : int;

                     instructControl         : byte;
                     loop                    : longint;

                     minimumDistance         : TT_F26dot6;
                     roundState              : int;

                     rp0,
                     rp1,
                     rp2                     : int;

                     scanControl             : boolean;
                     singleWidthCutIn        : TT_F26dot6;
                     singleWidthValue        : TT_F26dot6;
                   end;



  {* TrueType Table Directory type *}

  TTableDir = Record
                version     : FixedPoint;   {* should be $10000 *}
                numTables   : word;         {* Tables number    *}

                searchRange,           {* These parameters are only used  *}
                entrySelector,         {* for a dichotomy search in the   *}
                rangeShift     : Word; {* directory. We ignore them       *}
               end;

  {* The 'TableDir' is followed by 'numTables' TableDirEntries *}

  TTableDirEntry = Record
                     Tag      : array[0..3] of Char; {*        table type *}
                     CheckSum : Long;                {*    table Checksum *}
                     Offset   : Long;                {* Table file offset *}
                     Length   : Long;                {*      Table length *}
                    end;

  TTableDirEntries = array[0..100] of TTableDirEntry;
  PTableDirEntries = ^TTableDirEntries;

  {* "cmap" Table *}

  TCMapDir = Record
               Version          : UShort;  {* should be 0 *}
               count            : UShort;  {* number of entries *}
              end;

  {* The "cmap" is followed by cMapNum TCMapDirEntries *}

  TCMapDirEntry = Record
                    PlatformID,                   {* Windows = 3          *}
                    PlatformEncodingID : UShort;  {* Microsoft UGL = 1    *}
                    offset             : Long;    {*       offset         *}
                   end;

  TCMapEntries = array[1..100] of TCMapDirEntry;
  PCMapEntries = ^TCMapEntries;

  {* Il existe 4 formats distincts : 0,2,4 et 6
     Nous ne nous intresserons ici qu'au format 4 *}

  TCMap4 = Record
             Format,                {* doit tre 4 *}
             Length,                {* taille en octets *}
             Version,               {* numro de version, dbute  0 *}
             segCountX2,            {* nombre de segments * 2 *}

             SearchRange,           {* paramtres utiliss pour une *}
             EntrySelector,         {* recherche optimise *}
             RangeShift     : UShort;
            end;

  TCMap4Segment = Record
                    endCount,
                    startCount,
                    idDelta,
                    idRangeOffset : UShort;
                   end;


  TCmap4_1 = array[1..100] of ushort;
  PCMap4_1 = ^TCMap4_1;

  TCMap4Segments = record
                         endCount,
                         startCount,
                         idDelta,
                         idRangeOffset : PCmap4_1;
                   end;

  TGlyphIDarray = array[1..10000] of ushort;
  PGlyphIDarray = ^TGlyphIDarray;


  {* table "maxp" des 'Maximum Profiles' *}

  TMaxProfile = Record
                  Version                 : Fixed;
                  numGlyphs,
                  maxPoints,
                  maxContours,
                  maxCompositePoints,
                  maxCompositeContours,
                  maxZones,
                  maxTwilightPoints,
                  maxStorage,
                  maxFunctionDefs,
                  maxInstructionDefs,
                  maxStackElements,
                  maxSizeOfInstructions,
                  maxComponentElements,
                  maxComponentDepth       : UShort;
                end;


  {* table de tag "glyph" *}

  TPointRec = record
               x, y, flag : integer;
              end;

  TPoints = array[0..1023] of TPointRec;
  PPoints = ^TPoints;

  PGlyphContour = ^TGlyphContour;
  TGlyphContour = record
                   Start,
                   Finish : word;
                  end;

  PGlyphContours = ^TGlyphContours;
  TGlyphContours = array[0..1023] of TGlyphContour;

  PGlyph = ^TGlyph;
  TGlyph = Record
            numberOfContours,
            xMin,
            yMin,
            xMax,
            yMax,
            numberOfPoints : Short;

            Contours : PGlyphContours;
            Points   : PPoints;

            composite : boolean;           { ci je to composed glyph }
            comp_n:integer;                { z kolkych glyphov }
            comp_:array[1..5] of record    { z ktorych a kde su umiestnene }
                                     which:integer;
                                     x,y:short;
                                 end;
           end;

  TGlyphs = array[0..1000] of TGlyph;
  PGlyphs = ^TGlyphs;

  {* table de type "head" *}

  TLongDateTime = record
                    L1,
                    L2 : long;
                  end;

  THeader = Record
             TableVersion : FixedPoint;
             FontRevision : FixedPoint;

             CheckSumAdjust : Long;
             MagicNumber    : Long;

             Flags      : UShort;
             UnitsPerEM : UShort;

             Created  : TLongDateTime;
             Modified : TLongDateTime;

             xMin : Short;
             yMin : Short;
             xMax : Short;
             yMax : Short;

             MacStyle      : UShort;
             LowestRecPPEM : UShort;

             FontDirection     : Short;
             IndexToLocFormat  : Short;
             GlyphDataFormat   : Short;
            end;


  
  ThmtxItem = record
                    aw,lsb:word;
              end;

  PhmtxItems = ^ThmtxItems;
  ThmtxItems = array[0..500] of ThmtxItem;




{* Les tableaux apparatront dans la mmoire selon l'ordre des
   dclarations suivantes                                      *}

type
  (* "loca" tableau des indices dans Glyph *)
  TLoca = Record
            Size  : word;
            Table : PStorage;
          end;

  TLocas = array[0..1000] of TLoca;


Type { Composite Glyph Description }
    TCompositeGlyph = record
                            flags : UShort;
                            glyphIndex : UShort;

                            argument1,argument2 : Short;
                            arg1and2 : UShort;

                            scale,xscale,yscale,scale01,scale10 : tt_f2dot14;
                      end;

var
  Table_Dir          : TTableDir;
  Table_Dir_Entries  : PTableDirEntries;
  Num_TDE            : int;

  MaxProfile         : TMaxProfile;

  Font_Header        : ^THeader;
  Glyph_Locations    : ^TLoca;
  Glyphs             : ^TGlyphs;

  Num_Glyphs         : int;    (* Number of glyphs in current font file   *)

  GS        : TGraphicsState; (* Current Graphics State record  *)
                              (* NOTE : this is not a pointer ! *)

  CVT       : PShortArray;   (* Pointer to the current CVT *)
  CvtSize   : int;           (* Size of the current CVT    *)

  Scale1,               (* These values are used to scale quickly *)
  Scale2    : LongInt;  (* from EM to pixel coordinates           *)

  PointSize : TT_F26dot6;  (* The PointSize in 26.6 fixed point *)

  CMapDir : TCMapDir;
  CMapDirEntries : PCMapEntries;
  CMap4Table : TCMap4;
  CMap4Segments : TCMap4Segments;
  GlyphIDarray : PGlyphIDarray;

  composite:integer;
  comp:TCompositeGlyph;

  hhea_tab : record
                   sth:array[1..34] of byte;
                   hmtx_count : word;
             end;

  hmtx_tab:PhmtxItems;


Const HexTab:array[0..15] of char = '0123456789ABCDEF';


function   Open_TrueType_File( AName : String ) : boolean;
procedure  Close_TrueType_File;

function Load_TrueType_Tables : boolean;
function LookUp_TrueType_Table( ATag : string ) : int;
function Load_TrueType_Header : boolean;
function Load_TrueType_CVT : boolean;
function Load_TrueType_CMap : boolean;
function Load_TrueType_Locations : Boolean;

function Load_TrueType_MaxProfile : boolean;
function Load_TrueType_Glyphs : integer;
Function Load_New_Composed : boolean;
Function TrueType_Compose : boolean;

Function CMAP_glyph(unicode:word) : integer;

Procedure Load_hmtx;

function strupper(s:string):string;

implementation

uses TTMemory,
{$IFDEF OS2}
     TTFileM;
{$ELSE}
     TTFile;
{$ENDIF}

(*************************)
(*  Load_TrueType_Tables *)
(*************************)

function Load_TrueType_Tables : Boolean;
var
  T : LongInt;
  L : LongInt;
  max_offset,max_len,sum_len:longint;

begin
  Load_TrueType_Tables := False;

  if not Read_At_Font_File( 0, Table_Dir, sizeof(Table_Dir) ) then exit;

  Do32( Fixed( Table_Dir.Version ) );
  Do16( Table_Dir.NumTables );

{$IFDEF DEBUG}
{  Writeln('Version de rpertoire : ',Table_Dir.version/$10000);}
  Writeln('Nombre de Tables      : ',Table_Dir.numTables);

{$ENDIF}

  Num_TDE := Table_Dir.NumTables;

  L := sizeof(TTableDirEntry) * Num_TDE;

  if not Alloc( L, Pointer(Table_Dir_Entries) ) or
     not Read_Font_File( Table_Dir_Entries^, L ) then exit;

  max_offset:=0; max_len:=0;
  for t:=0 to Num_TDE-1 do with Table_Dir_Entries^[t] do
    begin
     CheckSum:=0;
     Do32( Offset );
     Do32( Length );

     If offset>max_offset then begin max_offset:=offset; max_len:=length; end;
   end;

  sum_len:=max_offset+max_len;

  Load_TrueType_Tables := True;
end;

(***************************)
(*  LookUp_TrueType_Table  *)
(***************************)

function LookUp_TrueType_Table( ATag : string ) : int;
var
  TAG : String[4];
  i   : int;
begin
  TAG[0] := #4;
  LookUp_TrueType_Table := -1;

  if Table_Dir_Entries = nil then exit;

  for i := 0 to Num_TDE-1 do
    begin

      move( Table_Dir_Entries^[i].Tag, Tag[1], 4 );

      if Tag = ATag then
        begin
          LookUp_TrueType_Table := i;
          exit;
        end

    end
end;

(**************************)
(*  Load_TrueType_Header  *)
(**************************)

function  Load_TrueType_Header : Boolean;
var
  i : int;
begin
  Load_TrueType_Header := False;

  i := LookUp_TrueType_Table('head');
  if i <= 0 then exit;

  if not Alloc( sizeof(THeader), Pointer(Font_Header) ) or

     not Read_At_FOnt_File( Table_Dir_Entries^[i].Offset,
                            Font_Header^, sizeof(THeader) )
      then exit;

  with Font_Header^ do
    begin

      Do16( word(IndexToLocFormat) );
      Do16( UnitsPerEM );
  {$IFDEF DEBUG}
      Writeln('Units per EM       : ',UnitsPerEM );
      Writeln('IndexToLocFormat   : ',IndexToLocFormat );
      Writeln('Glyphs number      : ',num_Glyphs );
  {$ENDIF}
    end;

  Load_TrueType_Header := True;
end;

(***************************)
(* Load_TrueType_Locations *)
(***************************)

function Load_TrueType_Locations : Boolean;
var
  i           : int;
  sz          : longint;
  t           : int;
  LongOffsets : int;
  locs        : PStorage;
  locs2       : PShortArray;
  Mrk         : TMarkRecord;

begin

  Load_TrueType_Locations := False;
  LongOffsets             := 0;

  if Font_Header = nil then
    if not Load_TrueType_Header then exit;

  LongOffsets :=  Font_Header^.IndexToLocFormat;

  (* default offsets format is short *)

  T := LookUp_TrueType_Table('loca');
  if T < 0 then exit;

  if not Alloc( sizeof(TLoca), Pointer(Glyph_Locations) ) then exit;

  if LongOffsets <> 0 then
    begin
      sz := Table_Dir_Entries^[T].Length shr 2;
      Glyph_Locations^.Size := sz;

      {$IFDEF DEBUG}
      Writeln('Glyph Slots # ( 32-bits offsets ) : ', sz );
      {$ENDIF}

      if not Alloc( 4*Sz, Pointer( Locs ) ) then exit;

      Glyph_Locations^.Table := locs;

      if not Read_At_Font_File( Table_Dir_Entries^[T].Offset,
                                Locs^[0], Sz*4 ) then exit;

      Do32s( locs^[0], sz );
    end
  else
    begin
      sz := Table_Dir_Entries^[T].Length shr 1;
      Glyph_Locations^.Size := Sz;

      {$IFDEF DEBUG}
      Writeln('Glyph Slots # ( 16-bits offsets ) : ', Sz );
      {$ENDIF}

      if not Alloc( 4*Sz, Pointer(locs) ) then exit;
      Mark( Mrk );
      if not Alloc( 2*Sz, Pointer(locs2)  ) then exit;

      Glyph_Locations^.Table := locs;

      if not Read_At_Font_File( Table_Dir_Entries^[T].Offset,
                                locs2^[0], 2*sz ) then exit;

      Do16s( locs2^[0], sz );
      for i := 0 to sz-1 do Locs^[i] := 2*longint( locs2^[i] );

      if not Release( Mrk ) then exit;
    end;

  Load_TrueType_Locations := True;
end;



function Load_TrueType_CVT : boolean;
var
  m : int;
begin
  Load_TrueType_CVT := False;
  m := LookUp_TrueType_Table('cvt ');
  if m<0 then exit;

  with Table_Dir_Entries^[m] do
   begin
     GetMem( CVT, Length );
     CvtSize := Length div sizeof(Short);
     if not Read_At_Font_File( Offset, CVT^, Length ) then exit;
     Do16s( CVT^, CvtSize );
   end;

  Load_TrueType_CVT := True;
end;

(******************************)
(*  Load_TrueType_MaxProfile  *)
(******************************)

function Load_TrueType_MaxProfile : boolean;
var
  m : int;
begin

  Load_TrueType_MaxProfile := False;

  m:=LookUp_TrueType_Table('maxp');
  if m<0 then exit;

  if not Read_At_Font_File( Table_Dir_Entries^[m].Offset,
                            MaxProfile, sizeof(MaxProfile) ) then exit;

  with MaxProfile do
   begin
    Do32( Version );
    Do16( numGlyphs );
    Do16( maxPoints );
    Do16( maxContours );
    Do16( maxCompositePoints );
    Do16( maxCompositeContours );
    Do16( maxZones );
    Do16( maxTwilightPoints );
    Do16( maxStorage );
    Do16( maxFunctionDefs );
    Do16( maxInstructionDefs );
    Do16( maxStackElements );
    Do16( maxSizeOfInstructions );
    Do16( maxComponentElements );
    Do16( maxCOmponentDepth );
   end;

  Num_Glyphs               := MaxProfile.NumGlyphs;
  Load_TrueType_MaxProfile := True;
end;

(******************************)
(*  Load_TrueType_Cmap        *)
(******************************)

function Load_TrueType_Cmap : boolean;
var
  count,unicode,i,m : int;
  j,k,file_index,size:longint;
  f:text;

begin

  Load_TrueType_Cmap := False;

  m:=LookUp_TrueType_Table('cmap');
  if m<0 then exit;

  if not Read_At_Font_File( Table_Dir_Entries^[m].Offset,
                            CMapDir, sizeof(CMapDir) ) then exit;

  with CMapDir do
   begin
    Do16( Version );
    Do16( Count );
   end;

  size:=sizeof(TCMapDirEntry)*CMapDir.count;
  If not Alloc(size,pointer(CMapDirEntries)) then exit;

  if not Read_At_Font_File( Table_Dir_Entries^[m].Offset+4,
                            CMapDirEntries^, size ) then exit;

  unicode:=0;
  For i:=1 to CMapDir.count do
   With CMapDirEntries^[i] do
   Begin
        Do16( PlatformID );
        Do16( PlatformEncodingID );
        Do32( offset );

        If ((PlatformID=3) and (PlatformEncodingID=1)) then
        Begin
             unicode:=i;
             break;
        End;
   End;

  If unicode=0 then exit;   { ziadna UNICODE tabulka }
  With CMapDirEntries^[unicode] do
  Begin
       If not Read_At_Font_File( Table_Dir_Entries^[m].offset+offset,
                                 CMap4Table, sizeof(TCMap4) ) then exit;

       Do16(Cmap4Table.format);
       Do16(Cmap4Table.length);
       Do16(Cmap4Table.version);

       Do16(Cmap4Table.segCountX2);
       Do16(Cmap4Table.searchRange);
       Do16(Cmap4Table.entrySelector);
       Do16(Cmap4Table.rangeShift);

       If Cmap4Table.format<>4 then exit;
  End;


  count:=CMap4Table.segCountX2 shr 1;
  size:=count*sizeof(ushort);

  If not Alloc(size,pointer(CMap4Segments.endCount)) then exit;
  If not Alloc(size,pointer(CMap4Segments.startCount)) then exit;
  If not Alloc(size,pointer(CMap4Segments.idDelta)) then exit;
  If not Alloc(size,pointer(CMap4Segments.idRangeOffset)) then exit;

  file_index:=Table_Dir_Entries^[m].offset + CMapDirEntries^[unicode].offset+14;

  If not Read_At_Font_File( file_index, CMap4Segments.endCount^, size ) then exit;
  inc(file_index,size+sizeof(ushort));

  If not Read_At_Font_File( file_index, CMap4Segments.startCount^, size ) then exit;
  inc(file_index,size);

  If not Read_At_Font_File( file_index, CMap4Segments.idDelta^, size ) then exit;
  inc(file_index,size);

  If not Read_At_Font_File( file_index, CMap4Segments.idRangeOffset^, size ) then exit;
  inc(file_index,size);
{}

  For i:=1 to count do
    With CMap4Segments do
    Begin
         Do16(endCount^[i]);
         Do16(startCount^[i]);
         Do16(idDelta^[i]);
         Do16(idRangeOffset^[i]);
    End;

  size:=CMap4Table.length-(16+4*size);

  If not Alloc(size,pointer(glyphIDarray)) then exit;
  If not Read_At_Font_File(file_index,glyphIDarray^,size) then exit;

  assign(f,'zoom.log');
  {$I-}
  rewrite(f);
  {$I+}
  If IOresult=0 then
  Begin
     writeln(f,'start':10,'end':10,'idDelta':10,'idRangeOffset':15);
     writeln(f,'---------------------------------------------------------------');

     For i:=1 to count do
     Begin
          with CMap4Segments do
            writeln(f,startCount^[i]:10,endCount^[i]:10,idDelta^[i]:10,idRangeOffset^[i]:10);
     End;

     writeln(f,'');
     writeln(f,'---------------------------------------------------------------');
     writeln(f,'GlyphIDArray');
     writeln(f,'---------------------------------------------------------------');
     writeln(f,'');

     j:=1;
     k:=0;

     write(f,k:4,':   ');

     For i:=1 to (size shr 1) do
     Begin
          Do16(glyphIDarray^[i]);
          write(f,glyphidarray^[i]:5);
          inc(j);
          if j>10 then begin writeln(f,''); inc(k,10); write(f,k:4,':   '); j:=1; end;
     End;

     write(f,#12);

     close(f);
  End;

  Load_TrueType_Cmap := True;
end;

(****************************************************************************)
(****************************************************************************)
(****************************************************************************)
(****************************************************************************)
function strupper(s:string):string;
Var i:byte;
    s1:string;
begin
     s1:='';
     for i:=1 to length(s) do s1:=s1+upcase(s[i]);
     strupper:=s1;
end;


Function New_Composed_Count : integer;
Var t:text;
    count:integer;
Begin
     If ((ParamCount=2) and (StrUpper(ParamStr(2))='-COMPOSE')) then
     Begin
          assign(t,'compose.ttc');
          reset(t);

          read(t,count);

          close(t);

          New_Composed_count:=count;
     End Else New_Composed_Count:=0;
End;

Function Load_New_Composed : boolean;
Var t:text;
    count,i,j,n:integer;

    a,x,y:integer;
    Gl:Tglyph;

Begin
     Load_New_composed:=false;

     If ((ParamCount=2) and (StrUpper(ParamStr(2))='-COMPOSE')) then
     Begin
          assign(t,'compose.ttc');
          reset(t);

          read(t,count);
          For i:=1 to count do
          Begin

               gl:=Glyphs^[Num_Glyphs];
               gl.comp_n:=0;
               gl.composite:=true;

               gl.xmax:=-16000;
               gl.ymax:=-16000;
               gl.xmin:=16000;
               gl.ymin:=16000;

               readln(t,n);
               For j:=1 to n do
               Begin
                    read(t,a,x,y);

                    inc(Gl.comp_n);
                    Gl.comp_[Gl.comp_n].which:=a;
                    Gl.comp_[Gl.comp_n].x:=x;
                    Gl.comp_[Gl.comp_n].y:=y;
               End;

               Glyphs^[Num_Glyphs]:=Gl;

               inc(Num_Glyphs);
          End;

          close(t);

          Load_New_composed:=true;
     End;
End;
(****************************************************************************)
(****************************************************************************)
(****************************************************************************)
(****************************************************************************)


(**************************)
(*  Load_TrueType_Glyphs  *)
(**************************)

function Load_TrueType_Glyphs : integer;
var
  sz, szc, szp : int;
  i, j, k, cnt : int;
  b, c         : byte;

  offset : longint;
  locs   : PStorage;

  GL  : TGlyph;
  Con : PGlyphContours;
  Pts : PPoints;

  a:array[1..1000] of byte;
  a_ind:integer;

  new_count:integer;

label
  Suite,
  Compose,
  Fin;

begin
  Load_TrueType_Glyphs:=0;

  i:=LookUp_TrueType_Table('glyf');
  if i<0 then exit;

  Offset:=Table_Dir_Entries^[i].Offset;

  if Glyph_Locations=NIL then
   if not Load_TrueType_Locations then exit;

  locs := Glyph_Locations^.Table;
  sz   := Glyph_Locations^.Size;

  New_count:=New_Composed_count;

  if not Alloc( Sizeof( TGlyph)*(Sz+New_count), Pointer(Glyphs) ) then exit;

  j:=0;

  for i:=0 to sz-1 do

   begin

    if not Read_At_Font_File( Offset+locs^[i],
                              GL, 5*sizeof(Integer) ) then goto Suite;

    Gl.composite:=false;
    Gl.comp_n:=0;

    (* INVALID OFFSET ??? *)

    Do16( Word( Gl.numberOfContours ) );
    If Gl.numberofcontours<0 then
    Begin
         write(i:4);
         inc(composite);
    End;
    Do16( Word( Gl.xMin ) ); Do16( Word( Gl.yMin ) );
    Do16( Word( Gl.xMax ) ); Do16( Word( Gl.yMax ) );

    {$IFDEF DEBUG}
    Writeln(' # of Contours : ',Gl.numberOfContours );
    Writeln(' xMin : ',Gl.xMin:4,'  xMax : ',Gl.xMax);
    Writeln(' yMin : ',Gl.yMin:4,'  yMax : ',Gl.yMax);
    Writeln('-');
    {$ENDIF}

    szc:=Gl.numberOfContours;

    if szc<0 then Goto Compose;

    if szc>MaxProfile.maxContours then
     begin
{$IFDEF DEBUG}
      Writeln('ERROR: Glyph ',i,' has ',szc,' contours > ',
               maxProfile.maxContours );
      readkey;
{$ENDIF}
      goto Suite;
     end;

    GetMem( Con, Sizeof(TGlyphContour)*szc );
    If Con=NIL then Goto Fin;

    Gl.Contours:=Con;
    Szp:=0;
    For k:=0 to szc-1 do
     begin

      {$IFDEF DEBUG}
      Write( szp,' ');
      {$ENDIF}

      Con^[k].Start:=Szp;
      Read_16( Short(Szp) );
      Con^[k].Finish:=Szp;
      inc(Szp);
     end;

    Gl.numberOfPoints:=szp;

    (* Sauter les instructions *)

    Read_16(Short(k));

    {$IFDEF DEBUG}
    Writeln('Instructions size : ',k);
    {$ENDIF}

    Skip_Font_File( k );
{    For a_ind:=1 to k do
    Begin
         read_8(a[a_ind]);
    End;                  }

    GetMem( Pts, sizeof(TPointRec)*szp );
    if pts=NIL then
     begin
      FreeMem( Con, SizeOf(TGlyphContour)*Szc );
      Goto Fin;
     end;

    Gl.Points:=Pts;

  (* Lecture des flags *)

    k:=0;
    while (k<szp) do
     begin

      Read_8(c);
      Pts^[k].flag := c;
      inc(k);

      if c and 8 <> 0 then
       begin
        Read_8( b );
        cnt := b;

        while (cnt>0) do
         begin
          Pts^[k].flag := c;
          inc( k );
          dec( cnt );
         end;
       end;
     end;

  (* Lecture des X *)

    for k:=0 to szp-1 do
     with pts^[k] do
      if flag and 2 <> 0 then
       begin
        Read_8( c );
        if flag and 16 <> 0 then x:=integer(c) else x:=-integer(c);
       end
      else
       if flag and 16 <> 0 then x:=0 else Read_16(x);

  (* Lecture des Y *)

    for k:=0 to szp-1 do
     with pts^[k] do
      if flag and 4 <> 0 then
       begin
        Read_8( c );
        if flag and 32 <> 0 then y:=integer(c) else y:=-integer(c);
       end
      else
       if flag and 32 <> 0 then y:=0 else Read_16(y);

  (* Ajustement des coordonnes relatives  absolues *)

    for k:=1 to szp-1 do with Pts^[k] do
     begin
      inc( x, Pts^[k-1].x );
      inc( y, Pts^[k-1].y );
     end;

    Glyphs^[j]:=Gl;
    inc(j);

    goto Suite;

 Compose:        { this is a composed glyph }

    With comp do
    Begin
      Repeat
         read_16u(flags);
         read_16u(glyphIndex);

         If (flags and COMP__ARG_1_AND_2_ARE_WORDS<>0) then
         Begin
              read_16(argument1);
              read_16(argument2);
         End
         Else
         Begin
              read_16u(arg1and2);
         End;

         If (flags and COMP__WE_HAVE_A_SCALE<>0) then
         Begin
              read_16(scale);
         End
         Else
          If (flags and COMP__WE_HAVE_AN_X_AND_Y_SCALE<>0) then
          Begin
               read_16(xscale);
               read_16(yscale);
          End
          Else
           If (flags and COMP__WE_HAVE_A_TWO_BY_TWO<>0) then
           Begin
                read_16(xscale);
                read_16(scale01);
                read_16(scale10);
                read_16(yscale);
           End;

         If (flags and COMP__ARGS_ARE_XY_VALUES<>0) then
         Begin
              inc(Gl.comp_n);
              Gl.comp_[Gl.comp_n].which:=glyphIndex;
              Gl.comp_[Gl.comp_n].x:=argument1;
              Gl.comp_[Gl.comp_n].y:=argument2;
         End;

      Until (flags and COMP__MORE_COMPONENTS=0);

      Gl.composite:=true;
      Glyphs^[j]:=gl;
      inc(j);
    End;   { with }

 Suite:
   end;

 Fin:
  Num_Glyphs           := j;
  Load_TrueType_Glyphs := j;
end;

Function I2Hex(v:byte):string;
Begin
     I2Hex:=HexTab[(v shr 4) and 15]+HexTab[v and 15];
End;

Function Memory(v:integer):string;
Var s:string;
Begin
     s:=I2Hex(Mem[Seg(v):Ofs(v)+1]);
     s:=s+' '+I2Hex(Mem[Seg(v):Ofs(v)]);

     Memory:=s;
End;

Function TrueType_Compose : boolean;
Var Gl : TGlyph;
    i,j,k,point_count,cont_count:integer;
    cont_done,Point_index,cont_index:integer;

    can_write:boolean;
    file_text:text;

Begin
     TrueType_Compose:=false;

     assign(file_text,'compose.xy');
     {$I-}
     rewrite(file_text);
     {$I+}
     If IOresult=0 then can_write:=true else can_write:=false;

     For i:=0 to Num_Glyphs-1 do
     Begin
          If Glyphs^[i].composite=false then continue;

          With Glyphs^[i] do
          Begin
               point_count:=0;
               cont_count:=0;
               For j:=1 to comp_n do
               Begin
                    inc(point_count,Glyphs^[comp_[j].which].NumberOfPoints);
                    inc(cont_count,Glyphs^[comp_[j].which].NumberOfContours);
               End;

               GetMem( Contours, Sizeof(TGlyphContour)*cont_count );
               GetMem( Points, sizeof(TPointRec)*point_count );

               point_index:=0;
               cont_index:=0;

               cont_done:=0;

               For j:=1 to comp_n do
               Begin
                    For k:=1 to Glyphs^[comp_[j].which].NumberOfContours do
                    Begin
                         inc(cont_index);
                         Contours^[cont_index-1].start:=Glyphs^[comp_[j].which].Contours^[k-1].start+cont_done;
                         Contours^[cont_index-1].finish:=Glyphs^[comp_[j].which].Contours^[k-1].finish+cont_done;
                    End;

                    inc(cont_done,Glyphs^[comp_[j].which].NumberOfPoints);

                    For k:=1 to Glyphs^[comp_[j].which].NumberOfPoints do
                    Begin
                         inc(point_index);
                         Points^[point_index-1].x := Glyphs^[comp_[j].which].Points^[k-1].x + comp_[j].x;
                         Points^[point_index-1].y := Glyphs^[comp_[j].which].Points^[k-1].y + comp_[j].y;
                         Points^[point_index-1].flag := Glyphs^[comp_[j].which].Points^[k-1].flag;
                    End;

                    If Xmin > Glyphs^[comp_[j].which].xmin+comp_[j].x then Xmin:=Glyphs^[comp_[j].which].xmin+comp_[j].x;
                    If Xmax < Glyphs^[comp_[j].which].xmax+comp_[j].x then Xmax:=Glyphs^[comp_[j].which].xmax+comp_[j].x;
                    If Ymin > Glyphs^[comp_[j].which].ymin+comp_[j].y then Ymin:=Glyphs^[comp_[j].which].Ymin+comp_[j].y;
                    If Ymax < Glyphs^[comp_[j].which].ymax+comp_[j].y then Ymax:=Glyphs^[comp_[j].which].Ymax+comp_[j].y;
               End;

               NumberOfContours:=cont_count;
               NumberOfPoints:=point_count;

               {protokol}
               If can_write then
               Begin
                  writeln(file_text,'Glyph number : ',i);
                  writeln(file_text,'Xmin : ',xmin,' = ',Memory(xmin));
                  writeln(file_text,'Ymin : ',ymin,' = ',Memory(ymin));
                  writeln(file_text,'Xmax : ',xmax,' = ',Memory(xmax));
                  writeln(file_text,'Ymax : ',ymax,' = ',Memory(ymax));
                  writeln(file_text,'');
               End;
          End;

     End;

     If can_write then close(file_text);

     TrueType_Compose:=true;
End;


function Open_TrueType_File( AName : String ) : boolean;
begin
  Open_TrueType_File := Open_Font_File( AName );
end;

procedure Close_TrueType_File;
begin
  Close_Font_File;
end;


Function CMAP_glyph(unicode:word) : integer;
Var segm:word;
    segcount:word;
    l:longint;
Begin

     segcount:=cmap4table.segcountx2 div 2;

     segm:=0;

     CMAP_glyph:=-1;

     With cmap4segments do
     Begin
          Repeat
                inc(segm);
          Until endcount^[segm]>=unicode;

          If unicode<startcount^[segm] then exit;

          If idRangeOffset^[segm]=0 then
          Begin
               l:=unicode;
               l:=l+IDdelta^[segm];
               CMAP_glyph:=l and 65535;
          End
          Else
          Begin
               l:=(idRangeOffset^[segm] div 2) + (unicode-startcount^[segm]) - (segcount-segm);
               l:=l+idDelta^[segm];
               CMAP_glyph:=glyphIDarray^[l and 65535];
          End;
     End;
End;


Procedure Load_hmtx;
Var i:integer;
Begin
     i:=LookUp_TrueType_Table('hhea');

     Read_At_Font_File( Table_Dir_Entries^[i].Offset,
                            hhea_tab, sizeof(hhea_tab) );

     Do16(hhea_tab.hmtx_count);


     i:=LookUp_TrueType_Table('hmtx');

     GetMem(hmtx_tab,sizeof(ThmtxItem)*hhea_tab.hmtx_count);

     Read_At_Font_File( Table_Dir_Entries^[i].Offset,
                            hmtx_tab^, sizeof(ThmtxItem)*hhea_tab.hmtx_count);

     For i:=1 to hhea_tab.hmtx_count do
     Begin
          Do16(hmtx_tab^[i-1].aw);
          Do16(hmtx_tab^[i-1].lsb);
     End;
End;


begin
  composite:=0;

  Num_TDE           := 0;
  Table_Dir_Entries := nil;

  Font_Header     := nil;
  Glyph_Locations := nil;
  Glyphs          := nil;

  Num_Glyphs := 0;

end.
