{**********************************************************}
{*                                                        *}
{* The following routines are inline assembly, they are   *}
{* thus processor and bitness specific. Replace them      *}
{* with your own if you want to port the TrueType Engine  *}


{**********************************************************}
{* Four operations to convert from Motorola to Intel      *}


Var Endian_w:array[0..10239] of word;

Procedure Do16(var S);
Var w:word;
Begin
  move(S,w,2);
  w:=swap(w);
  Move(w,S,2);
End;

Procedure Do16s(var S; Cnt:Int);
Var i:integer;
Begin
  Move(S,Endian_w,2*cnt);
  For i:=0 to cnt-1 do Endian_w[i]:=swap(Endian_w[i]);
  Move(Endian_w,S,2*cnt);
End;

Procedure Xchg(var a,b:word);
Var c:word;
Begin
  c:=a;
  a:=b;
  b:=c;
End;

Procedure Do32(var L);
Var ww:array[0..1] of word;
Begin
  Move(L,ww,4);
  ww[0]:=swap(ww[0]);
  ww[1]:=swap(ww[1]);
  xchg(ww[0],ww[1]);
  Move(ww,L,4);
End;

Procedure Do32s(var L; Cnt:Int);
Var i:integer;
Begin
  Move(L,Endian_W,4*cnt);

  For i:=0 to 2*cnt-1 do Endian_W[i]:=swap(Endian_W[i]);
  For i:=0 to cnt-1 do Xchg(Endian_W[2*i],Endian_W[2*i+1]);

  Move(Endian_W,L,4*cnt);
End;


{**********************************************************}
{* Calc A*B/C with Intermediate 64 bit precision          *}

function MulDiv( A, B, C : Int32 ) : Int32;
const
  Bit31 = 1 shl 31;
label
  OverFlow;
var
  r, q, aq, ar, s1, s2 : Int32;
  i                    : Int;
begin
  s1 := a and Bit31; if s1 <> 0 then a := -a;
  s2 := b and Bit31; if s2 <> 0 then b := -b;

  s1 := s1 xor s2;

  s2 := c and Bit31; if s2 <> 0 then c := -c;

  s1 := s1 xor s2;

  if s1 <> 0 then
    s2 := $80000001
  else
    s2 := $7FFFFFFF;

  if c = 0 then goto OverFlow;

  { A Useful shortcut : most of the calculations fall here }

  if (a or b) shr 16 = 0 then
  begin
    q := a*b div c;
    if s1 <> 0 then q := -q;
    MulDiv := q;
    exit;
  end;

  aq := a div c;
  ar := a mod c;

  r := 0;
  q := 0;

  for i := 1 to 32 do
  begin
    r := r shl 1;
    q := q shl 1;

    { if q < 0 then goto OverFlow; }

    if b < 0 then
    begin
      inc( r, ar );
      inc( q, aq );

      { if q < 0 then goto OverFlow; }
    end;

    b := b shl 1;

    while (r-c) >= 0 do
    begin
      dec( r, c );
      inc( q );
     {  if q < 0 then goto OverFlow; }
    end;

  end;

  if s1 <> 0 then q := -q;
  MulDiv := q;
  exit;

OverFlow:
  MulDiv := s2;

end;


(*
{**********************************************************}
{* 64 Bit Addition                                        *}

procedure Add64( var X, Y, Z : Int64 ); assembler;
asm
  les si,[X]

  mov ax,es:[ si ].word
  mov dx,es:[si+2].word
  mov bx,es:[si+4].word
  mov cx,es:[si+6].word

  les si,[Y]
  add ax,es:[ si ].word
  adc dx,es:[si+2].word
  add bx,es:[si+4].word
  adc cx,es:[si+6].word

  les si,[Z]
  mov es:[ si ].word,ax
  mov es:[si+2].word,dx
  mov es:[si+4].word,bx
  mov es:[si+6].word,cx
end;


{**********************************************************}
{* 64 Bit Substraction                                    *}

procedure Sub64( var X, Y, Z : Int64 ); assembler;
asm
  les si,[X]

  mov ax,es:[ si ].word
  mov dx,es:[si+2].word
  mov bx,es:[si+4].word
  mov cx,es:[si+6].word

  les si,[Y]
  sub ax,es:[ si ].word
  sbb dx,es:[si+2].word
  sub bx,es:[si+4].word
  sbb cx,es:[si+6].word

  les si,[Z]
  mov es:[ si ].word,ax
  mov es:[si+2].word,dx
  mov es:[si+4].word,bx
  mov es:[si+6].word,cx
end;


{**********************************************************}
{* Multiply two Int32 to an Int64                         *}

procedure MulTo64( X, Y : Int32; var Z : Int64 ); assembler;
asm
  les si,[Z]
  db $66; mov ax,[X].word
  db $66; imul [Y].word
  db $66; mov es:[si],ax
  db $66; mov es:[si+4],dx
end;


{**********************************************************}
{* Divide an Int64 by an Int32                            *}

function Div64by32( var X : Int64; Y : Int32 ) : Int32; assembler;
asm
  les si,[X]

  db $66; mov ax,es:[si]
  db $66; mov dx,es:[si+2]
  db $66; idiv [Y].word
end;


{**********************************************************}
{* MSB index ( return -1 for 0 )                          *}

function Order64( var Z : Int64 ) : int; assembler;
asm
  les si,[Z]
  db $66; mov ax,es:[si]
  db $66; mov dx,es:[si+4]
  mov bx,63
  db $66; mov cx,0; dw $8000

 @1:
  db $66; test dx,cx
  jnz @3
  dec bx
  db $66; ror cx,1
  jns @1

 @2:
  db $66; test ax,cx
  jnz @2
  dec bx
  db $66; ror cx,1
  jns @2

 @3:
  mov ax,bx
end;


{**********************************************************}
{* MSB index ( return -1 for 0 )                          *}

function Order32( Z : Int32 ) : int; assembler;
asm
  les si,[Z]
  db $66; mov ax,es:[si]
  mov bx,31
  db $66; mov cx,0; dw $8000

 @1:
  db $66; test ax,cx
  jnz @2
  dec bx
  db $66; ror cx,1
  jns @1

 @2:
  mov ax,bx
end;
*)


{* End of processor specific routines                     *}
{*                                                        *}
{**********************************************************}
