(*

 Prezerac TrueTypeFonts
------------------------------------------------------------------------------

  Ak je iba jeden parameter (meno suboru), tak sa prezeraju jednotlive znaky,
  ovladanie je +,-,u,j (zoom) a k,l,i,o (znaky).

  Ak je druhym parametrom "-export", tak sa vykona export do <meno.exp> podla
  tabulky v subore EXPORT.TTC

*)

program Zoom;

uses Crt, TTDisp, TTTypes, TTMemory, TTCalc, TTTables, Raster, IO_check;

{&PMTYPE NOVIO}


const
  Precis  = 64;
  Precis2 = Precis div 2;

  PrecisAux = 1024;

  Centre_X : int = 320;
  Centre_Y : int = 225;

  Profile_Buff_Size = 64000;

  font:array[0..9] of array[0..7] of byte =
         (
           ($1C,$22,$22,$22,$22,$22,$22,$1C),
           ($08,$18,$28,$08,$08,$08,$08,$08),
           ($1C,$22,$02,$02,$04,$08,$10,$3E),
           ($1C,$22,$02,$04,$02,$02,$22,$1C),
           ($08,$10,$20,$48,$7C,$08,$08,$08),
           ($3E,$20,$20,$3C,$02,$02,$22,$1C),
           ($1C,$22,$20,$3C,$22,$22,$22,$1C),
           ($3E,$22,$02,$02,$04,$04,$08,$08),
           ($1C,$22,$22,$1C,$22,$22,$22,$1C),
           ($1C,$22,$22,$1E,$02,$02,$22,$1C)
         );



var

  Font_Buffer : PStorage;

  curGlyphContours : PGlyphContours;

  num_pts : word;
  num_ctr : word;

  glyfArray : word;

  epts_ctr : PShortArray;

  xCoord : PStorage;
  yCoord : PStorage;
  Flag   : PByteArray;

  ymin, ymax, xmax, xmin, xsize : longint;
  res,  resB                    : int;

  resR : real;

  resX, resY : real;

  LastX, LastY : FixedPoint;

  numPoints, numContours : int;

  curGlyph        : ^TGlyph;
  curGlyphContour : PGlyphContour;

  Bit : TRasterBlock;

  yCur : integer;

  ScXMax, ScYMax,
  CntX, CntY : Integer;

  Rotation : int;  (* Angle modulo 1024 *)


  f:text;
  s,expname:string;
  charcode:char;
  j,k,unicode:word;
  glyph:integer;

  conv_what:array[1..500] of record
                                   c:byte;        { char code }
                                   uni:word;      { unicode }
                                   glyph:integer; { glyph }
                             end;
  count:word;
  baseline:integer;
  counter,last:integer;

  first_x,first_y,a,b,last_x,last_y:integer;

  lsb,width:integer;

var medved_n,medved_i : integer;
             medved_h : real;



Procedure InitRows;
var
  i: integer;
  P: Pointer;
begin

  Bit.rows  := 450;
  Bit.cols  := 80;
  Bit.width := 640;
  Bit.flow  := TTFlowDown;
  Bit.size  := 80*450;

  GetMem( Bit.buffer, Bit.size );
  if Bit.buffer = NIL then
   begin
    Writeln('ERREUR:InitRows:Not enough memory to allocate BitMap');
    halt(1);
   end;

  { Note that the render pool should be allocated from the font pool }
  { for vairous debugging reasons, and because we're still in alpha, }
  { we don't do it yet..                                             }

  GetMem( P, Profile_Buff_Size );
  if P=nil then
   begin
    writeln('ERREUR:InitRows:Not enough memory to allocate render pool');
    Halt(2);
   end;

  InitRasterizer( Bit, P, Profile_Buff_Size );

  FillChar( Bit.Buffer^, Bit.Size, 0 );
end;


Procedure ClearData;
var i: integer;
begin
  FillChar( Bit.Buffer^, Bit.Size, 0 );

  FreeMem( XCoord, SizeOf(FixedPoint)*numPoints );
  FreeMem( YCoord, SizeOf(FixedPoint)*numPoints );

  FreeMem( Flag, numPoints );
end;


Function LoadTrueTypeChar( idx : integer ) : boolean;
var
  off    : longint;
  x, y   : Real;
  i, szp : integer;
  j      : word;
  c, ct  : byte;
  Gl     : TGlyph;
  EM     : Word;
  CR, SR : Real;

begin
  LoadtrueTypeChar:=FALSE;
  if (idx<0) or (idx>=Num_Glyphs) then exit;

  CurGlyph := @Glyphs^[idx];
  Gl       := CurGlyph^;

  numPoints        := Gl.numberOfPoints;
  numContours      := Gl.numberOfContours;
  curGlyphContours := Gl.Contours;

  GetMem( XCoord, SizeOf(Fixed)*numPoints );
  GetMem( YCoord, SizeOf(Fixed)*numPoints );
  GetMem( Flag, numPoints );

  xMin := Gl.xMin;
  xMax := Gl.xMax;
  yMin := Gl.yMin;
  yMax := Gl.yMax;

  EM := Font_Header^.UnitsPerEM;

  dec( xMax, xMin );
  dec( yMax, yMin );

  dec ( res );
  resR := res/EM/2;

  xmax := trunc( xmax * resR + 0.5 );
  ymax := trunc( ymax * resR + 0.5 );

  CR := Cos( Rotation*Pi/512 );
  SR := Sin( Rotation*Pi/512 );

  for j:=0 to numPoints-1 do
   begin

    x := Gl.Points^[j].x * ( res / EM );
    y := Gl.Points^[j].y * ( res / EM );

    off := Trunc( Precis*( CR*(x-xmax) + SR*(y-ymax) ) );

    XCoord^[j] := Precis*Centre_X + off;
    (*
    XCoord^[j] := Precis*( Centre_X + off div Precis ) + Precis2;
    *)

    off := Trunc( Precis*( - SR*(x-xmax) + CR*(y-ymax) ) );

    YCoord^[j] := Precis*Centre_Y + off;
    (*
    YCoord^[j] := Precis*( Centre_Y + off div Precis ) + Precis2;
    *)

    Flag^[j] := Gl.Points^[j].flag and 1;
   end;

  inc ( res );
  resR := 1/res;

  xsize := ( xmax + 7 ) div 8;

  LoadTrueTypeChar:=TRUE;
end;


function ConvertRaster : boolean;
var
  B : Array[0..128] of Integer;
  i : integer;
  G : TGlyphRecord;
begin

  for i := 0 to numContours-1 do
    B[i] := CurGlyphContours^[i].Finish;

  G.Outlines  := numContours;
  G.OutStarts := @B;
  G.Points    := numPoints;
  G.XCoord    := XCoord;
  G.YCoord    := YCoord;
  G.Flag      := Flag;

  ConvertRaster := RenderGlyph ( G, res, res );
end;




var i: integer;
    heure,
    min1,
    min2,
    sec1,
    sec2,
    cent1,
    cent2  :
{$IFDEF OS2}
    longint;
{$ELSE}
    word;
{$ENDIF}

    C : Char;

    Filename : String;

label Fin;

var
  Fail : Int;


function i2s(c:integer):string;
Var s:string;
Begin
     Str(c,s);
     i2s:=s;
End;

procedure show(c:integer);
var p:integer;
    n:byte;
Begin
     p:=76;
     repeat
           n:=c mod 10;
           c:=c div 10;

           vio^[p+0*80]:=font[n][0];
           vio^[p+1*80]:=font[n][1];
           vio^[p+2*80]:=font[n][2];
           vio^[p+3*80]:=font[n][3];
           vio^[p+4*80]:=font[n][4];
           vio^[p+5*80]:=font[n][5];
           vio^[p+6*80]:=font[n][6];
           vio^[p+7*80]:=font[n][7];

           dec(p);
     until c=0;
End;

Procedure Kill(a:string);
Var f:file;
Begin
     assign(f,a);
     {$I-}
     reset(f);
     {$I+}
     If IOresult<>0 then exit;

     close(f);
     erase(f);
End;


begin
  GetMem       ( Font_Buffer, 64000 );
  Init_FontPool( Font_Buffer^, 64000 );

  curGlyphContours:=NIL;

  num_pts   :=0;
  num_ctr   :=0;

  xCoord  :=NIL;
  yCoord  :=NIL;
  Flag    :=NIL;

  If paramCount=0 then
   begin
    Writeln('Simple TrueType Glyphs viewer - part of the FreeType project' );
    Writeln;
    Writeln('Usage : ',paramStr(0),' FontName[.TTF] [-compose]');
    Halt(1);
   end;


  Filename := ParamStr(1);
  if Pos('.',FileName)=0 then FileName:=FileName+'.TTF';
  if not Open_TrueType_File(Filename ) then
   begin
    Writeln('Erreur, le fichier ',ParamStr(1),' n''a pu tre ouvert');
    Halt(1);
   end;

  res  := 350;
  resB := (res+7) div 8;

  Rotation := 0;

  Fail := 0;


  Load_TrueType_Tables;
  Load_TrueType_CMap;
  Load_TrueType_MaxProfile;


  composite:=0;
  writeln('Looking for composite glyphs...');

  if Load_TrueType_Glyphs=0 then
   begin
    Writeln('Problme lors du chargement des glyphes');
    Halt(1);
   end;


  Load_New_Composed;

  TrueType_Compose;

  writeln(#13#10'Found ',composite,' composite glyphs.');
  {readln;} delay(500);

(*******  vymazanie kontrolnych suborov  ************************************)
(**)  kill('compose.xy');                                                 (**)
(**)  kill('zoom.log');                                                   (**)
(****************************************************************************)

  If ((ParamCount=2) and (strupper(paramstr(2))='-EXPORT')) then
  Begin
       kill('compose.xy');
       kill('zoom.log');

       Load_hmtx;

       OpenText(f,'export.ttc',o_READ);

       readln(f,s);

       count:=0;

       writeln;
       writeln('Loading "EXPORT.TTC"...');
       While s<>'' do
       Begin
            If ((s[1]=';') and (s[2]=';')) then
            begin
                 writeln('REM: '+copy(s,3,255));
                 readln(f,s);
                 continue;
            end;

            If s[1]='`' then charcode:=s[2]
                        else charcode:=char(s2i(Copy(s,1,pos(' ',s)-1)));

            delete(s,1,pos(' ',s));

            If s[1]='$' then
            Begin
                 Delete(s,1,1);
                 unicode:=x2i(s);
            End
            Else Unicode:=s2i(s);

            glyph:=CMAP_glyph(unicode);
            If glyph=-1 then glyph:=0;

            writeln('CHAR: ',byte(charcode):3,' (',charcode,')   UNICODE: ',unicode:5,'   GLYPH: ',glyph:5);

            inc(count);
            conv_what[count].c:=byte(charcode);
            conv_what[count].uni:=unicode;
            conv_what[count].glyph:=glyph;

            readln(f,s);
       End;

       writeln;
       writeln('Converting ',count,' characters...');

       Close(f);


       expname:=paramstr(1);
       If pos('.',Copy(expname,length(expname)-3,4))=0 then
       Begin      { bez pripony }
            expname:=expname+'.mf';     { metafont }
       End
       Else
       Begin      { s priponou }
            expname:=copy(expname,1,length(expname)-4)+'.mf';
       End;

       opentext(f,expname,o_REWRITE);

       writeln(f,'mode_setup;');
       writeln(f,'');
       writeln(f,'u#:=1pt#;');
       writeln(f,'define_pixels(u);');
       writeln(f,'currenttransform:=currenttransform scaled u;');
       writeln(f,'');
       writeln(f,'primarydef v picxor w =');
       writeln(f,' begingroup');
       writeln(f,'  picture q; q=v+w; cull q keeping (1,1); q');
       writeln(f,' endgroup');
       writeln(f,'enddef;');
       writeln(f,'');
       writeln(f,'picture pom;');
       writeln(f,'def neg expr p = pom:=nullpicture; addto pom contour p.t_;');
       writeln(f,'                 currentpicture:=currentpicture picxor pom; enddef;');
       writeln(f,'def c(expr a,b,c,d) = .. controls (a,b) and (c,d) .. enddef;');
       writeln(f,'');

       baseline:=glyphs^[CMAP_glyph(67)].ymin;

       For i:=1 to count do
        With conv_what[i] do
        Begin
             lsb:=hmtx_tab^[glyph].lsb;
             width:=hmtx_tab^[glyph].aw;

             writeln(f,' ',c);
             writeln(f,' ',width*2,' ',
                           (glyphs^[glyph].ymax-glyphs^[glyph].ymin)*2,' ',
                           (glyphs^[glyph].ymin-baseline)*2);

             writeln(f,'  ',glyphs^[glyph].NumberOfContours);

             For j:=1 to glyphs^[glyph].NumberOfContours do
             Begin
                  counter:=0;
                  last:=0;

                  For k:=glyphs^[glyph].Contours^[j-1].start to glyphs^[glyph].Contours^[j-1].finish do
                  Begin
                       If glyphs^[glyph].Points^[k].flag and 1 = 1 then
                       Begin
                            If last=1 then inc(counter);
                            last:=1;
                       End
                       Else
                       Begin
                            inc(counter);
                            last:=2;
                       End;
                  End;

                  If last=1 then inc(counter);

                  writeln(f,'  ',counter);

                  last_x:=0;
                  last_y:=0;
                  first_x:=0;
                  first_y:=0;

                  last:=0;

                  For k:=glyphs^[glyph].Contours^[j-1].start to glyphs^[glyph].Contours^[j-1].finish do
                  Begin
                       If glyphs^[glyph].Points^[k].flag and 1 = 1 then
                       Begin
                            If last=0 then
                            Begin
                                 last_x:=(glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2;
                                 last_y:=(glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2;

                                 first_x:=last_x;
                                 first_y:=last_y;
                            End;

                            If last=1 then
                            Begin
                                 writeln(f,'   ',last_x,' ',last_y,' ',
                                                 last_x,' ',last_y,' ',
                                                 (glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2,' ',
                                                  (glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2,' ',
                                                 (glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2,' ',
                                                  (glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2);

                                 last_x:=(glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2;
                                 last_y:=(glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2;
                            End;

                            If last=2 then
                            Begin
                                 writeln(f,'   ',last_x,' ',last_y,' ',
                                                 (glyphs^[glyph].Points^[k-1].x-glyphs^[glyph].xmin+lsb)*2,' ',
                                                  (glyphs^[glyph].Points^[k-1].y-glyphs^[glyph].ymin)*2,' ',
                                                 (glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2,' ',
                                                  (glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2,' ',
                                                 (glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2,' ',
                                                  (glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2);

                                 last_x:=(glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2;
                                 last_y:=(glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2;
                            End;


                            last:=1;
                       End
                       Else
                       Begin
                            If last=0 then ;

                            If last=1 then ;

                            If last=2 then
                            Begin
                                 a:=( (glyphs^[glyph].Points^[k-1].x-glyphs^[glyph].xmin+lsb)*2
                                        + (glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2) div 2;
                                 b:=( (glyphs^[glyph].Points^[k-1].y-glyphs^[glyph].ymin)*2
                                        + (glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2) div 2;

                                 writeln(f,'   ',last_x,' ',last_y,' ',
                                                 (glyphs^[glyph].Points^[k-1].x-glyphs^[glyph].xmin+lsb)*2,' ',
                                                  (glyphs^[glyph].Points^[k-1].y-glyphs^[glyph].ymin)*2,' ',
                                                 a,' ',b,' ',
                                                 a,' ',b);

                                 last_x:=a;
                                 last_y:=b;
                            End;


                            last:=2;
                       End;
                  End;

                  If last=1 then
                  Begin
                       writeln(f,'   ',last_x,' ',last_y,' ',
                                       last_x,' ',last_y,' ',
                                       first_x,' ',first_y,' ',
                                       first_x,' ',first_y);
                  End;

                  If last=2 then
                  Begin
                       writeln(f,'   ',last_x,' ',last_y,' ',
                                       (glyphs^[glyph].Points^[k].x-glyphs^[glyph].xmin+lsb)*2,' ',
                                        (glyphs^[glyph].Points^[k].y-glyphs^[glyph].ymin)*2,' ',
                                       first_x,' ',first_y,' ',
                                       first_x,' ',first_y);
                  End;

             End;
             writeln(f,'');
        End;

       close(f);

       writeln('Conversion done.');
  End
  Else
  Begin
       InitRows;

       SetGraphScreen;

       I   := Num_Glyphs-1;

       Repeat

         if LoadtrueTypeChar(i) then

           if ConvertRaster then

             Display( Bit.Buffer^, 450, 80 )

           else
             inc( Fail );

         Show(i);

         C:=Readkey;
         Case C of

          #27 : goto Fin;
          { ESC Key }

          'x' : Rotation := ( Rotation-1  ) and 1023;
          'c' : Rotation := ( Rotation+1  ) and 1023;
          'v' : Rotation := ( Rotation-16 ) and 1023;
          'b' : Rotation := ( Rotation+16 ) and 1023;

          '+' : if res < 1040 then inc( res, 10 ) else res := 1050;
          '-' : if res > 11 then dec( res, 10 ) else res := 1;

          'i' : if i > 10 then dec( i, 10 ) else i := 0;
          'o' : if i < num_glyphs-11 then inc( i, 10 )else i := num_glyphs-1;

          'k' : if i > 0 then dec(i);
          'l' : if i < num_glyphs-1 then inc(i);

          'u' : if res > 0 then dec(res);
          'j' : if res < 450 then inc(res);

          ' ' : i:=0;

         end;

         ClearData;

       Until false;

      Fin:
       RestoreScreen;

  End;


  Close_TrueType_File;

  CorrectExit;

end.

